@@ -8,11 +8,11 @@ In addition, code highlighting is supported when editing Python code files,like
 
 一款使用tkinter编写的文本编辑器, 支持编辑文本文件、二进制文件、改变字体大小。
 支持ansi、gbk、utf-8等编码。编辑二进制文件时, 文件内容以转义序列形式显示。
-支持查找、替换、改变字体大小; 且支持撤销、重做;
+支持查找、替换、改变字体大小; 且支持撤销、重做; 支持将文件拖入窗口
 编辑python代码文件时, 支持代码高亮显示, 类似IDLE。
 
 作者:qfcy (七分诚意)
-版本:1.2.6
+版本:1.2.8.04
 """
 import sys,os,time
 from tkinter import *
@@ -26,17 +26,17 @@ try:
     from idlelib.percolator import Percolator
 except ImportError:
     ColorDelegator=Percolator=None
-
+try:import windnd
+except ImportError:windnd=None
 
 __email__="3416445406@qq.com"
 __author__="七分诚意 qq:3076711200 邮箱:%s"%__email__
-__version__="1.2.6"
+__version__="1.2.8.04"
 
-#_REPLACE_CHARS=['\x00','\r']
 def view_hex(bytes):
     result=''
     for i in range(0,len(bytes)):
-        result+= hex(bytes[i])[2:] + ' '
+        result+= hex(bytes[i])[2:].zfill(2) + ' '
         if (i+1) % 4 == 0:result+='\n'
     return result
 
@@ -59,12 +59,21 @@ def to_bytes(escape_str):
     except SyntaxError:
         return eval("b'''"+escape_str+"'''")
 
+def bell_(widget=None):
+    try: 
+        import winsound
+        winsound.PlaySound('.',winsound.SND_ASYNC)
+    except (ImportError, RuntimeError):
+        if widget is not None:widget.bell()
+
+def handle(err,parent=None):
+    # showinfo()中,parent参数指定消息框的父窗口
+    msgbox.showinfo("错误",type(err).__name__+': '+str(err),parent=parent)
+
 class SearchDialog(Toplevel):
     #查找对话框
     instances=[]
     def __init__(self,master):
-        if not isinstance(master,Editor):
-            raise TypeError("The master must be an Editor object, not %r."%(type(master).__name__))
         self.master=master
         self.coding=self.master.coding.get()
         cls=self.__class__
@@ -83,7 +92,7 @@ class SearchDialog(Toplevel):
         frame.pack(side=RIGHT,fill=Y)
         inputbox=Frame(self)
         Label(inputbox,text="查找内容:").pack(side=LEFT)
-        self.keyword=StringVar()
+        self.keyword=StringVar(self.master)
         keyword=ttk.Entry(inputbox,textvariable=self.keyword)
         keyword.pack(side=LEFT,expand=True,fill=X)
         keyword.bind("<Key-Return>",self.search)
@@ -94,13 +103,13 @@ class SearchDialog(Toplevel):
         options.pack(fill=X)
     def create_options(self,master):
         Label(master,text="选项: ").pack(side=LEFT)
-        self.use_regexpr=IntVar()
+        self.use_regexpr=IntVar(self.master)
         ttk.Checkbutton(master,text="使用正则表达式",variable=self.use_regexpr)\
         .pack(side=LEFT)
-        self.match_case=IntVar()
+        self.match_case=IntVar(self.master)
         ttk.Checkbutton(master,text="区分大小写",variable=self.match_case)\
         .pack(side=LEFT)
-        self.use_escape_char=IntVar()
+        self.use_escape_char=IntVar(self.master)
         self.use_escape_char.set(self.master.isbinary)
         ttk.Checkbutton(master,text="转义字符",variable=self.use_escape_char)\
         .pack(side=LEFT)
@@ -110,7 +119,9 @@ class SearchDialog(Toplevel):
         key=self.keyword.get()
         if not key:return
         if self.use_escape_char.get():
-            key=str(to_bytes(key),encoding=self.coding)
+            try:key=str(to_bytes(key),encoding=self.coding)
+            except Exception as err:
+                handle(err,parent=self);return
         text.tag_remove("sel","1.0",END)
         pos=text.search(key,INSERT,END,
                         regexp=self.use_regexpr.get(),
@@ -121,14 +132,9 @@ class SearchDialog(Toplevel):
             if mark:self.mark_text(pos,newpos)
             return pos,newpos
         elif bell: # 未找到
-            try: # -*****- 1.2.5版更新
-                import winsound
-                winsound.PlaySound('.',winsound.SND_ASYNC)
-            except (ImportError, RuntimeError):
-                self.bell()
+            bell_(widget=self)
     def mark_text(self,start_pos,end_pos):
         text=self.master.contents
-        #text.tag_config("found",background="#8888ff")
         text.tag_add("sel", start_pos,end_pos)
         lines=text.get('1.0',END)[:-1].count(os.linesep) + 1
         lineno=int(start_pos.split('.')[0])
@@ -153,7 +159,7 @@ class ReplaceDialog(SearchDialog):
 
         inputbox=Frame(self)
         Label(inputbox,text="查找内容:").pack(side=LEFT)
-        self.keyword=StringVar()
+        self.keyword=StringVar(self.master)
         keyword=ttk.Entry(inputbox,textvariable=self.keyword)
         keyword.pack(side=LEFT,expand=True,fill=X)
         keyword.focus_force()
@@ -161,7 +167,7 @@ class ReplaceDialog(SearchDialog):
 
         replace=Frame(self)
         Label(replace,text="替换为:  ").pack(side=LEFT)
-        self.text_to_replace=StringVar()
+        self.text_to_replace=StringVar(self.master)
         replace_text=ttk.Entry(replace,textvariable=self.text_to_replace)
         replace_text.pack(side=LEFT,expand=True,fill=X)
         replace_text.bind("<Key-Return>",self.replace)
@@ -174,6 +180,7 @@ class ReplaceDialog(SearchDialog):
         text=self.master.contents
         result=self.search(mark=False,bell=bell)
         if not result:return -1 #-1标志已无文本可替换
+        self.master.text_change()
         pos,newpos=result
         newtext=self.text_to_replace.get()
         text.delete(pos,newpos)
@@ -197,6 +204,7 @@ class Editor(Tk):
     windows=[]
     def __init__(self,filename=""):
         super().__init__()
+        self.withdraw() # 暂时隐藏窗口,避免调用create_widgets()时窗口闪烁
         self.title(self.TITLE)
         self.bind("<Key>",self.window_onkey)
         self.bind("<FocusIn>",self.focus)
@@ -204,18 +212,17 @@ class Editor(Tk):
         self.protocol("WM_DELETE_WINDOW",self.ask_for_save)
 
         self.isbinary=self.file_changed=False
-        self.bin_data=self.charmap=None
         self.colorobj=self._codefilter=None
-        self.coding=StringVar()
+        self.coding=StringVar(self)
         self.coding.set(self.NORMAL_CODING)
         Editor.windows.append(self)
 
         self.load_icon()
         self.create_widgets()
-        self.update()
+        self.wm_deiconify();self.update() # wm_deiconfy恢复被隐藏的窗口
+        if windnd:windnd.hook_dropfiles(self,func=self.onfiledrag);self.drag_files=[]
         self.filename=''
         if filename:
-            #self.filenamebox.insert(END,filename)
             self.load(filename)
     def load_icon(self):
         for path in sys.path + [os.path.split(sys.executable)[0]]: # 用于Py2exe
@@ -224,11 +231,13 @@ class Editor(Tk):
             except TclError:pass
             else:break
     def create_widgets(self):
-        "创建控件"
+        # 创建控件
         self.statusbar=Frame(self)
         self.statusbar.pack(side=BOTTOM,fill=X)
         self.status=Label(self.statusbar,justify=RIGHT)
         self.status.pack(side=RIGHT)
+        self.bin_data=ScrolledText(self.statusbar,width=6,height=6)
+        self.charmap=ScrolledText(self.statusbar,width=14,height=5)
 
         frame=Frame(self)
         frame.pack(side=TOP,fill=X)
@@ -241,28 +250,33 @@ class Editor(Tk):
 
         Label(frame,text="编码:").pack(side=LEFT)
         coding=ttk.Combobox(frame,textvariable=self.coding)
+        def tip(event):
+            self.msg['text']='重新打开或保存即可生效'
+            self.msg.after(2500,clear)
+        def clear():self.msg['text']=''
+        coding.bind('<<ComboboxSelected>>',tip)
         coding["value"]=self.encodings
         coding.pack(side=LEFT)
+        self.msg=Label(frame)
+        self.msg.pack(side=LEFT)
 
         self.contents=ScrolledText(self,undo=True,width=75,height=24,
                                    font=(self.NORMAL_FONT,self.NORMAL_FONTSIZE,"normal"))
         self.contents.pack(expand=True,fill=BOTH)
         self.contents.bind("<Key>",self.text_change)
         self.contents.bind("<B1-ButtonRelease>",self.update_status)
+        order = self.contents.bindtags() # 修复无法获取选定的文本的bug
+        self.contents.bindtags((order[1], order[0])+order[2:])
         self.update_offset()
 
         self.create_menu()
     def create_binarytools(self):
         if self.isbinary:
-            if not self.bin_data:
-                self.bin_data=ScrolledText(self.statusbar,width=6,height=6)
-            if not self.charmap:
-                self.charmap=ScrolledText(self.statusbar,width=14,height=5)
             self.bin_data.pack(side=LEFT,expand=True,fill=BOTH)
             self.charmap.pack(fill=Y)
             self.status.pack_forget()
             self.status.pack(fill=X)
-        else:
+        else: # 隐藏工具
             if self.bin_data:
                 self.bin_data.pack_forget()
             if self.charmap:
@@ -273,10 +287,10 @@ class Editor(Tk):
         filemenu=Menu(self,tearoff=False)
         filemenu.add_command(label="新建",
                              command=self.new,accelerator="Ctrl+N")
+        filemenu.add_command(label="新建二进制文件",command=self.new_binary)
         filemenu.add_command(label="打开",
                              command=self.open,accelerator="Ctrl+O")
-        filemenu.add_command(label="打开二进制文件",
-                             command=self.open_as_binary)
+        filemenu.add_command(label="打开二进制文件",command=self.open_as_binary)
         filemenu.add_command(label="保存",
                              command=self.save,accelerator="Ctrl+S")
         filemenu.add_command(label="另存为",command=self.save_as)
@@ -284,20 +298,26 @@ class Editor(Tk):
         filemenu.add_command(label="退出",command=self.ask_for_save)
 
         self.editmenu=Menu(self.contents,tearoff=False)
-        master=self.contents
+        master = self.contents
         self.editmenu.add_command(label="剪切  ",
-                         command=lambda:self.text_change()==master.event_generate("<<Cut>>"))
+                         command=lambda:self.text_change()\
+                                ==master.event_generate("<<Cut>>"))
         self.editmenu.add_command(label="复制  ",
                          command=lambda:master.event_generate("<<Copy>>"))
         self.editmenu.add_command(label="粘贴  ",
-                         command=lambda:self.text_change()==master.event_generate("<<Paste>>"))
+                         command=lambda:self.text_change()\
+                                ==master.event_generate("<<Paste>>"))
         self.editmenu.add_separator()
         self.editmenu.add_command(label="查找",accelerator="Ctrl+F",
                                   command=lambda:self.show_dialog(SearchDialog))
         self.editmenu.add_command(label="替换",
                                   command=lambda:self.show_dialog(ReplaceDialog))
-        self.editmenu.add_separator()
 
+        view=Menu(self.contents,tearoff=False)
+        self.is_autowrap=IntVar(self.contents) # 是否自动换行
+        self.is_autowrap.set(1)
+        view.add_checkbutton(label="自动换行", command=self.set_wrap,
+                             variable=self.is_autowrap)
         fontsize=Menu(self.contents,tearoff=False)
         fontsize.add_command(label="增大字体   ",accelerator='Ctrl+ "+"',
                              command=self.increase_font)
@@ -310,17 +330,28 @@ class Editor(Tk):
                 self.set_fontsize(index)
             fontsize.add_command(label=self.FONTSIZES[i],command=resize)
 
-        self.editmenu.add_cascade(label="字体",menu=fontsize)
         self.contents.bind("<Button-3>",
                     lambda event:self.editmenu.post(event.x_root,event.y_root))
-
+        view.add_cascade(label="字体",menu=fontsize)
 
         helpmenu=Menu(self,tearoff=False)
         helpmenu.add_command(label="关于",command=self.about)
 
         menu.add_cascade(label="文件",menu=filemenu)
         menu.add_cascade(label="编辑",menu=self.editmenu)
+        menu.add_cascade(label="查看",menu=view)
         menu.add_cascade(label="帮助",menu=helpmenu)
+
+        popup1=Menu(self.bin_data,tearoff=False)
+        popup1.add_command(
+            label="复制",command=lambda:self.bin_data.event_generate("<<Copy>>"))
+        popup2=Menu(self.charmap,tearoff=False)
+        popup2.add_command(
+            label="复制",command=lambda:self.charmap.event_generate("<<Copy>>"))
+        self.bin_data.bind("<Button-3>",
+                    lambda event:popup1.post(event.x_root,event.y_root))
+        self.charmap.bind("<Button-3>",
+                    lambda event:popup2.post(event.x_root,event.y_root))
         # 显示菜单
         self.config(menu=menu)
 
@@ -330,51 +361,49 @@ class Editor(Tk):
         for window in dialog_type.instances:
             if window.master is self:
                 window.focus_force()
-                return
+                return # 不重复显示对话框
         dialog_type(self).show()
 
     def set_fontsize(self,index):
         newsize=self.FONTSIZES[index]
         self.contents["font"]=(self.NORMAL_FONT,newsize,"normal")
     def increase_font(self):
-        "增大字体"
+        # 增大字体
         fontsize=int(self.contents["font"].split()[1])
         index=self.FONTSIZES.index(fontsize)+1
         if 0<=index<len(self.FONTSIZES): self.set_fontsize(index)
     def decrease_font(self):
-        "减小字体"
+        # 减小字体
         fontsize=int(self.contents["font"].split()[1])
         index=self.FONTSIZES.index(fontsize)-1
         if 0<=index<len(self.FONTSIZES): self.set_fontsize(index)
-
+    def set_wrap(self):
+        if self.is_autowrap.get():
+            self.contents['wrap'] = CHAR
+        else:
+            self.contents['wrap'] = NONE
+          # 注意:不需要此行代码
+##        self.is_autowrap.set(int(not self.is_autowrap.get()))
 
     def window_onkey(self,event):
-        if event.state==4:#如果按下Ctrl键
-            if event.keysym=='o':#按下Ctrl+O键
+        # 如果按下Ctrl键
+        if event.state in (4,6,12,14,36,38,44,46): # 适应多种按键情况(Num,Caps,Scroll)
+            key=event.keysym.lower()
+            if key=='o':#按下Ctrl+O键
                 self.open()
-            elif event.keysym=='s':#Ctrl+S键
+            elif key=='s':#Ctrl+S键
                 self.save()
-            elif event.keysym=='z':#Ctrl+Z
-                try:self.contents.edit_undo()
-                except TclError:self.bell()
-                self.text_change()
-            elif event.keysym=='y':
-                try:self.contents.edit_redo()
-                except TclError:self.bell()
-                self.text_change()
-            elif event.keysym=='f':
+            elif key=='f':
                 self.show_dialog(SearchDialog)
-            elif event.keysym=='equal':#Ctrl+ "+" 增大字体
+            elif key=='equal':#Ctrl+ "+" 增大字体
                 self.increase_font()
-            elif event.keysym=='minus':#Ctrl+ "-" 减小字体
+            elif key=='minus':#Ctrl+ "-" 减小字体
                 self.decrease_font()
     def focus(self,event):
-        #当窗口获得或失去焦点时,调用此函数
+        #当窗口获得或失去焦点时,调用此函数, 用于使对话框置于主窗体前
         for window in SearchDialog.instances + ReplaceDialog.instances:
             if window.master is self:
                 if event.type==EventType.FocusIn:
-                    if window.wm_state()=="iconic":
-                        window.attributes("-toolwindow",True)
                     window.attributes("-topmost",True)
                     if not window.wm_state()=="normal":
                         window.deiconify()
@@ -388,28 +417,34 @@ class Editor(Tk):
 
     def text_change(self,event=None):
         self.file_changed=True
-        self.update_status()
+        self.update_status();self.change_title()
     def update_status(self,event=None):
         if self.isbinary:
             try:
                 selected=self.contents.get(SEL_FIRST,SEL_LAST)
                 data=to_bytes(selected)
-                text=str(data,encoding=self.coding.get(),
-                         errors="backslashreplace")
+                try:text=str(data,encoding=self.coding.get(),
+                             errors="backslashreplace")
+                except TypeError:
+                    # 忽略Python 3.4的bug: don't know how to handle
+                    # UnicodeDecodeError in error callback
+                    text=str(data,encoding=self.coding.get(),
+                             errors="replace")
                 self.bin_data.delete("1.0",END)
                 self.bin_data.insert(INSERT,text)
                 self.charmap.delete("1.0",END)
                 self.charmap.insert(INSERT,view_hex(data))
                 self.status["text"]="选区长度: %d (Bytes)"%len(data)
-            except (TclError,TypeError,SyntaxError): #未选取内容, 或格式不正确
+            except (TclError,SyntaxError): #未选取内容, 或格式不正确
                 self.bin_data.delete("1.0",END)
                 self.charmap.delete("1.0",END)
-        self.update_offset()
+                self.update_offset()
+        else:self.update_offset()
     def update_offset(self,event=None):
         if self.isbinary:
             prev=self.contents.get("1.0",INSERT)
             try:
-                data=to_bytes(prev) # to_bytes_long(prev,master=self)
+                data=to_bytes(prev)
             except SyntaxError:
                 sep='\\'
                 # self.update()
@@ -424,22 +459,35 @@ class Editor(Tk):
     def new(cls):
         window=cls()
         window.focus_force()
-    def open(self):
+    @classmethod
+    def new_binary(cls):
+        window=cls()
+        window.isbinary=True
+        window.create_binarytools()
+        window.change_title()
+        window.change_mode()
+        window.contents.edit_reset()
+        window.focus_force()
+    def open(self,filename=None):
         #加载一个文件
         if self.ask_for_save(quit=False)==0:return
-        filename=dialog.askopenfilename(master=self,title='打开',
-                                        filetypes=self.FILETYPES)
+        if not filename:
+            filename=dialog.askopenfilename(master=self,title='打开',
+                                initialdir=os.path.split(self.filename)[0],
+                                filetypes=self.FILETYPES)
         self.load(filename)
     def open_as_binary(self):
+        if self.ask_for_save(quit=False)==0:return
         filename=dialog.askopenfilename(master=self,title='打开二进制文件',
-                                        filetypes=self.FILETYPES)
+                                initialdir=os.path.split(self.filename)[0],
+                                filetypes=self.FILETYPES)
         self.load(filename,binary=True)
     def load(self,filename,binary=False):
         # 加载文件(主要部分)
         if not filename.strip():
             return
         self.isbinary=binary
-        if os.path.isfile(filename):
+        try:
             data=self._load_data(filename)
             if data==0:return
             self.filename=filename
@@ -453,11 +501,12 @@ class Editor(Tk):
                     except TclError:self.contents.insert(INSERT,' ')
             self.contents.mark_set(INSERT,"1.0")
             self.create_binarytools()
+            self.file_changed=False
             self.change_title()
             self.change_mode()
-            self.contents.edit_reset() # -*****- 1.2.5版更新 !!!!!!!!!!!!!!!!
+            self.contents.edit_reset() # -*****- 1.2.5版更新 !!!!!!!
             self.contents.focus_force()
-        else:msgbox.showinfo("Error","文件未找到:"+repr(filename))
+        except Exception as err:handle(err,parent=self)
     def _load_data(self,filename):
         f=open(filename,"rb")#打开文件
         if self.isbinary:
@@ -470,7 +519,7 @@ class Editor(Tk):
             except UnicodeDecodeError:
                 f.seek(0)
                 result=msgbox.askyesnocancel("PyNotepad","""%s编码无法解码此文件,
-是否使用二进制模式打开?"""%self.coding.get())
+是否使用二进制模式打开?"""%self.coding.get(),parent=self)
                 if result:
                     self.isbinary=True
                     data=to_escape_str(f.read())
@@ -479,11 +528,14 @@ class Editor(Tk):
                     data=str(f.read(),encoding=self.coding.get(),errors="replace")
                 else:
                     return 0 # 表示取消
-        return data
-    def change_title(self,running=False):
-        newtitle="PyNotepad - "+os.path.split(self.filename)[1]+\
-                  (" (二进制模式)" if self.isbinary else '')+\
-                  (" (运行)" if running else '')
+            # return data.replace('\r','')
+            return data
+    def change_title(self):
+        file = os.path.split(self.filename)[1] or "未命名"
+        newtitle="PyNotepad - "+ file +\
+                  (" (二进制模式)" if self.isbinary else '')
+        if self.file_changed:
+            newtitle="*%s*"%newtitle
         self.title(newtitle)
     def change_mode(self):
         if ColorDelegator:
@@ -500,7 +552,8 @@ class Editor(Tk):
         if self.file_changed: ## and self.filenamebox.get():
             retval=msgbox.askyesnocancel("文件尚未保存",
                               "是否保存{}的更改?".format(
-                                  os.path.split(self.filename)[1] or "当前文件"))
+                            os.path.split(self.filename)[1] or "当前文件"),
+                                parent=self)
             if not retval is None:
                 if retval==True:
                     # 是
@@ -519,30 +572,45 @@ class Editor(Tk):
     def save(self):
         #保存文件
         if not self.filename:
-            self.filename=dialog.asksaveasfilename(master=self.master,
+            self.filename=dialog.asksaveasfilename(master=self,
+                    initialdir=os.path.split(self.filename)[0],
                     filetypes=self.FILETYPES)
         filename=self.filename
         if filename.strip():
-            text=self.contents.get('1.0', END)[:-1] # [:-1]: 去除换行符
-            if self.isbinary:
-                data=to_bytes(text)
-            else:data=bytes(text,encoding=self.coding.get())
-            file=open(filename, 'wb')
-            file.write(data)
-            file.close()
-            self.filename=filename
-            self.file_changed=False
+            try:
+                text=self.contents.get('1.0', END)[:-1] # [:-1]: 去除换行符
+                if self.isbinary:
+                    data=to_bytes(text)
+                else:
+                    data=bytes(text,encoding=self.coding.get(),errors='replace')
+                with open(filename, 'wb') as f:
+                    f.write(data)
+                self.filename=filename
+                self.file_changed=False
+            except Exception as err:handle(err,parent=self)
             self.change_title()
             self.change_mode()
         else:
             return 0 # 0表示cancel
     def save_as(self):
-        self.filename=dialog.asksaveasfilename(master=self.master,
+        filename=dialog.asksaveasfilename(master=self,
+                    initialdir=os.path.split(self.filename)[0],
                     filetypes=self.FILETYPES)
-        self.save()
+        if filename:
+            self.filename=filename
+            self.save()
 
     def about(self):
-        msgbox.showinfo("关于",__doc__+"\n作者: "+__author__)
+        msgbox.showinfo("关于",__doc__+"\n作者: "+__author__,parent=self)
+    def onfiledrag(self,files):
+        self.drag_files=files
+        self.after(50,self.onfiledrag2)
+    def onfiledrag2(self):
+        if not self.contents.get('1.0',END).strip() and not self.file_changed:
+            self.open(self.drag_files[0].decode('gbk'))
+            del self.drag_files[0]
+        for item in self.drag_files:
+            Editor(item.decode('gbk'))
 
 def main():
     if len(sys.argv)>1:
